// TelemetryDevice.h : header for abstract TelemetryDevice class
//

#pragma once

#include "decl.h"
#include "TelemetryTag.h"
#include "Engine/Engine.h"
#include "Utils/Semaphore.h"
#include "Utils/SerialPort.h"
#include "Utils/Time.h"

// The global dll info structure.
extern AFX_EXTENSION_MODULE TelemetryDeviceDLL;

typedef bool (* LPFNEXTERNALFORMATTER)(TData& data, int iGroup, TTime tRef);
typedef LPFNEXTERNALFORMATTER (* LPFN_INSTALL_CUSTOM_FORMATTER)(CString& strName);

#define MISSED_TX_TIMES_REGISTRY_KEY (_T("System\\Sutron\\Retransmit\\"))

/////////////////////////////////////////////////////////////////////////////
// TTelemetryDevice
//
class TELEMETRY_DECL TTelemetryDevice : public TPropertyContainer
{
   // 
   // Interface
   // 
   public:

   // Constructor and destructor. 
   TTelemetryDevice();
   virtual ~TTelemetryDevice();
   
   void LoadCustomFormatter();
   void UnloadCustomFormatter();

   // Device names
   virtual CString GetAbbrev() = 0;                   // e.g., "IR"
   virtual CString GetMoniker();                      // e.g., "IR2:"
   virtual CString GetName() = 0;                     // e.g., "Iridium"
   virtual CString GetFullName();                     // e.g., "Iridium COM2:"
   
   // Setup methods. 
   virtual CString GetSetupTag();                     // e.g., "Iridium2"
   virtual bool InitSetup();

   // Baseline device states 
   virtual void SetState(const CString& strState);
   virtual CString GetState();
   virtual CString GetShortStatus();
   virtual CString GetLongStatus();

   // Methods to invoke a public start and stop.
   virtual void SignalStart();
   virtual void SignalStop();
   virtual bool IsRunning() {return m_strState != _T("Stopping") && m_strState != _T("Stopped");}
   virtual bool WaitForStop(DWORD dwWait = INFINITE);
   virtual bool IsEnabled() {return propEnabled.AsInteger() && propPort.AsInteger(); };
   virtual bool IsStopped() {return WaitForSingleObject(m_hStopEvent, 0) == WAIT_OBJECT_0; };
   virtual bool IsAlertEnabled();
   bool WaitForRecStop(DWORD dwWait = INFINITE);
   
   // Get time of next scheduled transmission.
   TTime NextScheduledTxTime(TTime tRef = TTime::GetCurrentTime());
   
   // Missed transmission helpers
   void LoadMissedTxTimes();
   void ClearMissedTxTimes();
   void CheckClearMissedTxTimes();
   TTime GetMissedTxTime(int idx = 0);
   void SetMissedTxTime(TTime t, int idx = 0);
   TTime GetOldestMissedTxTime();
   TTime GetBufferStartTime();

   // Formatting interface. 
   enum EFormats {ePseudobinaryB, ePseudobinaryBNonInt, ePseudobinaryC, ePseudobinaryD, eSHEF, eSHEFFIX, 
      eNIFC, eNFRDS, EHandar, eSSP, eText, eASCIIColumn, eCSV, eBUFR, eNumFormats};
   static TSelectionInfo siFormats;
   EFormats GetFormat();
   void SetFormat(EFormats fmt);
   bool IsFormatInterleaved() {EFormats format =  GetFormat(); return IsFormatInterleaved(format);}
   bool IsFormatInterleaved(EFormats format) 
      {return format == TTelemetryDevice::ePseudobinaryB || format == TTelemetryDevice::eASCIIColumn;}
   bool FormatForcedTx(TData& data);
   bool FormatSelfTimed(TData& data, TTime tRef = TTime::GetCurrentTime(), int iGroup = 1);
   bool FormatForced(TData& data, TTime tRef = TTime::GetCurrentTime());
   bool FormatAlarm(TData& data, TTime tRef = TTime::GetCurrentTime(), ETagStates eCheckFlag=eTriggered, ETagStates eSetFlag=eFormatted);
   bool AllContentIs(ETxContent content);
   int GetNumVals() { return propNumVals.AsInteger(); };

   // Transmit selector for alarm data.
   enum EAlarmDataSelection {eSensorsInAlarm, eAllSensors, eNumAlarmDataSelections};
   static TSelectionInfo siAlarmDataSelection;
   EAlarmDataSelection GetAlarmDataSelection();
   void SetAlarmDataSelection(EAlarmDataSelection sel);
   virtual void SignalAlert(bool bIntoAlarm);
   virtual void ResetAlert();
   virtual bool InAlert();
   virtual void ClearTransmittedTagFlags();
   virtual void EnableAlert();
   virtual void DisableAlert();

   // Signal a forced transmission. Leave strMsg NULL to send self-time format
   virtual void SignalForcedTx(LPCTSTR strMsg = NULL) {}

   // Methods to create, maintain, and delete telemetry tags and tag list. 
   virtual void AddTag(TTelemetryTag* pTag);
   virtual void RemoveTag(TTelemetryTag* pTag);
   virtual void SortTags() {tags.SortTags();}
   virtual bool SelectTags(CWnd* pParent);

   // Common properties
   TProperty propPort;
   TProperty propEnabled;
   TProperty propTxTime;
   TProperty propTxInterval;
   TProperty propTxRetransmit;
   TProperty propTxFormat;
   TProperty propNumVals;
   TProperty propAlarmData;
   TProperty propOriginatingCenter;
   TProperty propDataCategory;
   TProperty propDataSubcategory;

   TTelemetryTags tags;

   // 
   // Implementation
   // 
   protected:
   friend class TTelemetryDeviceMenu;
   friend UINT DeviceThread(LPVOID pParam);

   virtual void Create();              // Call in derived constructor for proper context
   void AddCommonProperties();
   virtual void Destroy();             // Call in derived destructor for proper context
   
   // Override to add custom properties during Create()
   virtual void AddCustomProperties() = 0;

   // These methods are called from the device thread, which is started by SignalStart(). 
   virtual bool OnStart() = 0;
   virtual void OnRun() = 0;
   virtual bool OnStop() = 0;

   // Low level formatting methods
   virtual bool FormatAllSensors(TData& data, int iGroup, TTime tRef = TTime::GetCurrentTime());
   virtual bool FormatAlarmTriggers(TData& data, int iGroup, ETagStates eCheckFlag=eTriggered, ETagStates eSetFlag=eFormatted);

   // Attributes
   bool m_bRecordingOn;                // State of recording when computing next task
   CString m_strState;                 // Processing state: stopped, started, etc. 
   CRITICAL_SECTION m_csState;         // Lock for state variable.
   HANDLE m_hStopEvent;                // Event used to signal stop. 
   HANDLE m_hAlertEvent;               // Event used to signal an alert
   HANDLE m_hAckAlertEvent;            // Event used to signal when alerts are acknowledged
   bool m_bIntoAlarm;                  // Track whether last alarm signal was into alarm.
   bool m_bInAlert;                    // Currently processing an alarm transmission (alert)
   HANDLE m_hForcedTxEvent;            // Event used to signal a forced transmission
   CString m_strForcedTxMsg;           // Message to send in forced tx, if any
   CWinThread* m_pThread;              // Pointer to device thread object. 
   CSerialPort m_port;                 // Device comm port

   LPFNEXTERNALFORMATTER m_lpfnExternalFormatter;
   HINSTANCE m_hInstCustomFormatter;

   // Oldest missed tx time for each server defined. 
   TArray<TTime> m_arMissedTxTimes;
};

extern "C" TELEMETRY_DECL bool IsTelemetrySupported(int Port);

// If msg is NULL or empty string, a self-timed tx is formatted and sent
extern "C" TELEMETRY_DECL bool SignalTelemetryForcedTx(int Port = 0, LPCTSTR strMsg = NULL);

extern "C" TELEMETRY_DECL bool SignalTelemetryAlert(int Port=0, bool bIntoAlarm=true);
extern "C" TELEMETRY_DECL bool ResetTelemetryAlert(int Port=0);
extern "C" TELEMETRY_DECL void EnableTelemetryAlert();
extern "C" TELEMETRY_DECL void DisableTelemetryAlert();
extern "C" TELEMETRY_DECL bool IsTelemetryAlertEnabled();
extern "C" TELEMETRY_DECL bool IsTelemetryInAlert(int Port=0);

// Call only when recording is off or transitioning to off
extern "C" TELEMETRY_DECL void ClearMissedTxTimes();

// The global device list
extern "C" TELEMETRY_DECL TObArray<TTelemetryDevice> TelemetryDeviceList;

// Manages read-write access to TelemetryDeviceList
extern "C" TELEMETRY_DECL TRWLock TelemetryDeviceLock;

class TEngineStopHook : public CAlarmMgr
{
public:
   virtual void OnEngineStop();
};
